"""
Unit Tests For Decorators
~~~~~~~~~~~~~~~~~~~~~~~~~
"""
import argh


def test_aliases():
    @argh.aliases("one", "two")
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_ALIASES)
    assert attr == ("one", "two")


def test_arg():
    @argh.arg("foo", help="help", nargs="+")
    @argh.arg("--bar", default=1)
    def func():
        pass

    attrs = getattr(func, argh.constants.ATTR_ARGS)
    assert attrs == [
        dict(option_strings=("foo",), help="help", nargs="+"),
        dict(option_strings=("--bar",), default=1),
    ]


def test_named():
    @argh.named("new-name")
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_NAME)
    assert attr == "new-name"


def test_wrap_errors():
    @argh.wrap_errors([KeyError, ValueError])
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_WRAPPED_EXCEPTIONS)
    assert attr == [KeyError, ValueError]


def test_wrap_errors_processor():
    @argh.wrap_errors(processor="STUB")
    def func():
        pass

    attr = getattr(func, argh.constants.ATTR_WRAPPED_EXCEPTIONS_PROCESSOR)
    assert attr == "STUB"


def test_expects_obj():
    @argh.expects_obj
    def func(args):
        pass

    attr = getattr(func, argh.constants.ATTR_EXPECTS_NAMESPACE_OBJECT)
    assert attr is True



"""
Argh
~~~~
"""
#
#  Copyright © 2010—2023 Andrey Mikhaylenko and contributors
#
#  This file is part of Argh.
#
#  Argh is free software under terms of the GNU Lesser
#  General Public License version 3 (LGPLv3) as published by the Free
#  Software Foundation. See the file README.rst for copying conditions.
#
from .assembling import (
    SUPPORTS_ALIASES,
    add_commands,
    add_subcommands,
    set_default_command,
)
from .decorators import aliases, arg, expects_obj, named, wrap_errors
from .dispatching import (
    PARSER_FORMATTER,
    ArghNamespace,
    EntryPoint,
    dispatch,
    dispatch_command,
    dispatch_commands,
)
from .exceptions import AssemblingError, CommandError, DispatchingError
from .helpers import ArghParser
from .interaction import confirm, safe_input

__all__ = (
    "SUPPORTS_ALIASES",
    "add_commands",
    "add_subcommands",
    "set_default_command",
    "aliases",
    "arg",
    "expects_obj",
    "named",
    "wrap_errors",
    "PARSER_FORMATTER",
    "ArghNamespace",
    "EntryPoint",
    "dispatch",
    "dispatch_command",
    "dispatch_commands",
    "AssemblingError",
    "CommandError",
    "DispatchingError",
    "ArghParser",
    "confirm",
    "safe_input",
)



"""
Unit Tests For Utility Functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
import functools

from argh.utils import get_arg_spec, unindent


def function(x, y=0):
    return


def decorated(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        print("Wrapping function call")
        return func(*args, **kwargs)

    return wrapped


def _assert_spec(func, **overrides):
    spec = get_arg_spec(func)

    defaults = {
        "args": ["x", "y"],
        "varargs": None,
        "varkw": None,
        "defaults": (0,),
        "kwonlyargs": [],
        "annotations": {},
    }

    for k in defaults:
        actual = getattr(spec, k)
        expected = overrides[k] if k in overrides else defaults[k]
        assert actual == expected


def test_get_arg_spec__plain_func():
    _assert_spec(function)


def test_get_arg_spec__decorated_func():
    def d(_f):
        return _f

    decorated = d(function)

    _assert_spec(decorated)


def test_get_arg_spec__wrapped():
    wrapped = decorated(function)
    _assert_spec(wrapped)


def test_get_arg_spec__wrapped_nested():
    wrapped = decorated(decorated(function))
    _assert_spec(wrapped)


def test_get_arg_spec__wrapped_complex():
    def wrapper_deco(outer_arg):
        def _outer(func):
            @functools.wraps(func)
            def _inner(*args, **kwargs):
                return func(*args, **kwargs)

            return _inner

        return _outer

    wrapped = wrapper_deco(5)(function)

    _assert_spec(wrapped)


def test_get_arg_spec__static_method():
    class C:
        @staticmethod
        def func(x, y=0):
            return x

    _assert_spec(C.func)


def test_get_arg_spec__method():
    class C:
        def func(self, x, y=0):
            return x

    _assert_spec(C.func, args=["self", "x", "y"])


def test_util_unindent():
    "Self-test the unindent() helper function"

    # target case
    one = """
    a
     b
      c
    """
    assert (
        unindent(one)
        == """
a
 b
  c
"""
    )

    # edge case: lack of indentation on first non-empty line
    two = """
a
  b
    c
"""

    assert unindent(two) == two

    # edge case: unexpectedly unindented in between
    three = """
    a
b
    c
    """

    assert (
        unindent(three)
        == """
a
b
c
"""
    )



"""
Unit Tests For Autocompletion
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""
from unittest.mock import patch

import argh


@patch("argh.completion.COMPLETION_ENABLED", True)
@patch("argh.completion.argcomplete")
def test_enabled(mock_argcomplete):
    parser = argh.ArghParser()

    parser.autocomplete()

    mock_argcomplete.autocomplete.assert_called_with(parser)


@patch("argh.completion.COMPLETION_ENABLED", False)
@patch("argh.completion.argcomplete")
@patch("argh.completion.os.getenv")
@patch("argh.completion.logger")
def test_disabled_without_bash(mock_logger, mock_getenv, mock_argcomplete):
    mock_getenv.return_value = "/bin/sh"
    parser = argh.ArghParser()

    parser.autocomplete()

    mock_argcomplete.assert_not_called()
    mock_logger.debug.assert_not_called()


@patch("argh.completion.COMPLETION_ENABLED", False)
@patch("argh.completion.argcomplete")
@patch("argh.completion.os.getenv")
@patch("argh.completion.logger")
def test_disabled_with_bash(mock_logger, mock_getenv, mock_argcomplete):
    mock_getenv.return_value = "/bin/bash"
    parser = argh.ArghParser()

    parser.autocomplete()

    mock_argcomplete.assert_not_called()
    mock_getenv.assert_called_with("SHELL", "")
    mock_logger.debug.assert_called_with(
        "Bash completion is not available. Please install argcomplete."
    )



"""
Integration Tests
~~~~~~~~~~~~~~~~~
"""
import argparse
import re
import sys
import unittest.mock as mock

import pytest

import argh
from argh.exceptions import AssemblingError
from argh.utils import unindent

from .base import CmdResult as R
from .base import DebugArghParser, get_usage_string, run

if sys.version_info < (3, 10):
    HELP_OPTIONS_LABEL = "optional arguments"
else:
    HELP_OPTIONS_LABEL = "options"


def test_set_default_command_integration():
    def cmd(foo=1):
        return foo

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "") == R(out="1\n", err="")
    assert run(p, "--foo 2") == R(out="2\n", err="")
    assert run(p, "--help", exit=True) == 0


def test_set_default_command_integration_merging():
    @argh.arg("--foo", help="bar")
    def cmd(foo=1):
        return foo

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "") == R(out="1\n", err="")
    assert run(p, "--foo 2") == R(out="2\n", err="")
    assert "bar" in p.format_help()


#
# Function can be added to parser as is
#


def test_simple_function_no_args():
    def cmd():
        yield 1

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "") == R(out="1\n", err="")


def test_simple_function_positional():
    def cmd(x):
        yield x

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "", exit=True) == "the following arguments are required: x"
    assert run(p, "foo") == R(out="foo\n", err="")


def test_simple_function_defaults():
    def cmd(x="foo"):
        yield x

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "") == R(out="foo\n", err="")
    assert run(p, "bar", exit=True) == "unrecognized arguments: bar"
    assert run(p, "--x bar") == R(out="bar\n", err="")


def test_simple_function_varargs():
    def func(*file_paths):
        # `paths` is the single positional argument with nargs='*'
        yield ", ".join(file_paths)

    p = DebugArghParser()
    p.set_default_command(func)

    assert run(p, "") == R(out="\n", err="")
    assert run(p, "foo") == R(out="foo\n", err="")
    assert run(p, "foo bar") == R(out="foo, bar\n", err="")


def test_simple_function_kwargs():
    @argh.arg("foo")
    @argh.arg("--bar")
    def cmd(**kwargs):
        # `kwargs` contain all arguments not fitting ArgSpec.args and .varargs.
        # if ArgSpec.keywords in None, all @arg()'s will have to fit ArgSpec.args
        for k in sorted(kwargs):
            yield "{0}: {1}".format(k, kwargs[k])

    p = DebugArghParser()
    p.set_default_command(cmd)

    message = "the following arguments are required: foo"
    assert run(p, "", exit=True) == message
    assert run(p, "hello") == R(out="bar: None\nfoo: hello\n", err="")
    assert run(p, "--bar 123", exit=True) == message
    assert run(p, "hello --bar 123") == R(out="bar: 123\nfoo: hello\n", err="")


@pytest.mark.xfail
def test_simple_function_multiple():
    raise NotImplementedError


@pytest.mark.xfail
def test_simple_function_nested():
    raise NotImplementedError


@pytest.mark.xfail
def test_class_method_as_command():
    raise NotImplementedError


def test_all_specs_in_one():
    @argh.arg("foo")
    @argh.arg("--bar")
    @argh.arg("fox")
    @argh.arg("--baz")
    def cmd(foo, bar=1, *args, **kwargs):
        yield "foo: {0}".format(foo)
        yield "bar: {0}".format(bar)
        yield "*args: {0}".format(args)
        for k in sorted(kwargs):
            yield "** {0}: {1}".format(k, kwargs[k])

    p = DebugArghParser()
    p.set_default_command(cmd)

    # 1) bar=1 is treated as --bar so positionals from @arg that go **kwargs
    #    will still have higher priority than bar.
    # 2) *args, a positional with nargs='*', sits between two required
    #    positionals (foo and fox), so it gets nothing.
    assert run(p, "one two") == R(
        out="foo: one\n" "bar: 1\n" "*args: ()\n" "** baz: None\n" "** fox: two\n",
        err="",
    )

    # two required positionals (foo and fox) get an argument each and one extra
    # is left; therefore the middle one is given to *args.
    assert run(p, "one two three") == R(
        out="foo: one\n"
        "bar: 1\n"
        "*args: ('two',)\n"
        "** baz: None\n"
        "** fox: three\n",
        err="",
    )

    # two required positionals (foo and fox) get an argument each and two extra
    # are left; both are given to *args (it's greedy).
    assert run(p, "one two three four") == R(
        out="foo: one\n"
        "bar: 1\n"
        "*args: ('two', 'three')\n"
        "** baz: None\n"
        "** fox: four\n",
        err="",
    )


def test_arg_merged():
    """@arg merges into function signature."""

    @argh.arg("my", help="a moose once bit my sister")
    @argh.arg("-b", "--brain", help="i am made entirely of wood")
    def gumby(my, brain=None):
        return my, brain, "hurts"

    p = DebugArghParser("PROG")
    p.set_default_command(gumby)
    help_msg = p.format_help()

    assert "a moose once bit my sister" in help_msg
    assert "i am made entirely of wood" in help_msg


def test_arg_mismatch_positional():
    """An `@arg('positional')` must match function signature."""

    @argh.arg("bogus-argument")
    def confuse_a_cat(vet, funny_things=123):
        return vet, funny_things

    p = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        p.set_default_command(confuse_a_cat)

    msg = (
        "confuse_a_cat: argument bogus-argument does not fit "
        "function signature: vet, -f/--funny-things"
    )
    assert msg in str(excinfo.value)


def test_arg_mismatch_flag():
    """An `@arg('--flag')` must match function signature."""

    @argh.arg("--bogus-argument")
    def confuse_a_cat(vet, funny_things=123):
        return vet, funny_things

    p = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        p.set_default_command(confuse_a_cat)

    msg = (
        "confuse_a_cat: argument --bogus-argument does not fit "
        "function signature: vet, -f/--funny-things"
    )
    assert msg in str(excinfo.value)


def test_arg_mismatch_positional_vs_flag():
    """An `@arg('arg')` must match a positional arg in function signature."""

    @argh.arg("foo")
    def func(foo=123):
        return foo

    p = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        p.set_default_command(func)

    msg = (
        'func: argument "foo" declared as optional (in function signature)'
        " and positional (via decorator)"
    )
    assert msg in str(excinfo.value)


def test_arg_mismatch_flag_vs_positional():
    """An `@arg('--flag')` must match a keyword in function signature."""

    @argh.arg("--foo")
    def func(foo):
        return foo

    p = DebugArghParser("PROG")
    with pytest.raises(AssemblingError) as excinfo:
        p.set_default_command(func)

    msg = (
        'func: argument "foo" declared as positional (in function signature)'
        " and optional (via decorator)"
    )
    assert msg in str(excinfo.value)


class TestErrorWrapping:
    def _get_parrot(self):
        def parrot(dead=False):
            if dead:
                raise ValueError("this parrot is no more")
            else:
                return "beautiful plumage"

        return parrot

    def test_error_raised(self):
        parrot = self._get_parrot()

        p = DebugArghParser()
        p.set_default_command(parrot)

        assert run(p, "") == R("beautiful plumage\n", "")
        with pytest.raises(ValueError) as excinfo:
            run(p, "--dead")
        assert re.match("this parrot is no more", str(excinfo.value))

    def test_error_wrapped(self):
        parrot = self._get_parrot()
        wrapped_parrot = argh.wrap_errors([ValueError])(parrot)

        p = DebugArghParser()
        p.set_default_command(wrapped_parrot)

        assert run(p, "") == R("beautiful plumage\n", "")
        assert run(p, "--dead") == R("", "ValueError: this parrot is no more\n", exit=1)

    def test_processor(self):
        parrot = self._get_parrot()
        wrapped_parrot = argh.wrap_errors([ValueError])(parrot)

        def failure(err):
            return "ERR: " + str(err) + "!"

        processed_parrot = argh.wrap_errors(processor=failure)(wrapped_parrot)

        p = argh.ArghParser()
        p.set_default_command(processed_parrot)

        assert run(p, "--dead") == R("", "ERR: this parrot is no more!\n", exit=1)

    def test_stderr_vs_stdout(self):
        @argh.wrap_errors([KeyError])
        def func(key):
            db = {"a": 1}
            return db[key]

        p = argh.ArghParser()
        p.set_default_command(func)

        assert run(p, "a") == R(out="1\n", err="")
        assert run(p, "b") == R(out="", err="KeyError: 'b'\n", exit=1)


def test_argv():
    def echo(text):
        return "you said {0}".format(text)

    p = DebugArghParser()
    p.add_commands([echo])

    _argv = sys.argv

    sys.argv = sys.argv[:1] + ["echo", "hi there"]
    assert run(p, None) == R("you said hi there\n", "")

    sys.argv = _argv


def test_commands_not_defined():
    p = DebugArghParser()

    assert run(p, "", {"raw_output": True}).out == p.format_usage()
    assert run(p, "").out == p.format_usage() + "\n"

    assert "unrecognized arguments" in run(p, "foo", exit=True)
    assert "unrecognized arguments" in run(p, "--foo", exit=True)


def test_command_not_chosen():
    def cmd(args):
        return 1

    p = DebugArghParser()
    p.add_commands([cmd])

    # returns a help message and doesn't exit
    assert "usage:" in run(p, "").out


def test_invalid_choice():
    def cmd(args):
        return 1

    # root level command

    p = DebugArghParser()
    p.add_commands([cmd])

    assert "invalid choice" in run(p, "bar", exit=True)

    # exits with an informative error
    assert run(p, "--bar", exit=True) == "unrecognized arguments: --bar"

    # nested command

    p = DebugArghParser()
    p.add_commands([cmd], group_name="nest")

    assert "invalid choice" in run(p, "nest bar", exit=True)

    # exits with an informative error
    assert run(p, "nest --bar", exit=True) == "unrecognized arguments: --bar"


def test_unrecognized_arguments():
    def cmd():
        return 1

    # single-command parser

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "--bar", exit=True) == "unrecognized arguments: --bar"
    assert run(p, "bar", exit=True) == "unrecognized arguments: bar"

    # multi-command parser

    p = DebugArghParser()
    p.add_commands([cmd])

    assert run(p, "cmd --bar", exit=True) == "unrecognized arguments: --bar"
    assert run(p, "cmd bar", exit=True) == "unrecognized arguments: bar"


def test_echo():
    "A simple command is resolved to a function."

    def echo(text):
        return "you said {0}".format(text)

    p = DebugArghParser()
    p.add_commands([echo])

    assert run(p, "echo foo") == R(out="you said foo\n", err="")


def test_bool_action():
    "Action `store_true`/`store_false` is inferred from default value."

    def parrot(dead=False):
        return "this parrot is no more" if dead else "beautiful plumage"

    p = DebugArghParser()
    p.add_commands([parrot])

    assert run(p, "parrot").out == "beautiful plumage\n"
    assert run(p, "parrot --dead").out == "this parrot is no more\n"


def test_bare_group_name():
    "A command can be resolved to a function, not a group_name."

    def hello():
        return "hello world"

    p = DebugArghParser()
    p.add_commands([hello], group_name="greet")

    # without arguments

    # returns a help message and doesn't exit
    assert "usage:" in run(p, "greet").out

    # with an argument

    # exits with an informative error
    message = "unrecognized arguments: --name=world"
    assert run(p, "greet --name=world", exit=True) == message


# TODO: remove in v.0.30
def test_bare_group_name__deprecated_arg():
    "A command can be resolved to a function, not a namespace."

    def hello():
        return "hello world"

    p = DebugArghParser()
    with pytest.warns(
        DeprecationWarning,
        match=r"Argument `namespace` is deprecated .+ it will be removed in Argh 0\.30",
    ):
        p.add_commands([hello], namespace="greet")

    # without arguments

    # returns a help message and doesn't exit
    assert "usage:" in run(p, "greet").out

    # with an argument

    # exits with an informative error
    message = "unrecognized arguments: --name=world"
    assert run(p, "greet --name=world", exit=True) == message


def test_function_under_group_name():
    "A subcommand is resolved to a function."

    def hello(name="world"):
        return "Hello {0}!".format(name or "world")

    def howdy(buddy):
        return "Howdy {0}?".format(buddy)

    p = DebugArghParser()
    p.add_commands([hello, howdy], group_name="greet")

    assert run(p, "greet hello").out == "Hello world!\n"
    assert run(p, "greet hello --name=John").out == "Hello John!\n"
    assert run(p, "greet hello John", exit=True) == "unrecognized arguments: John"

    # exits with an informative error
    message = "the following arguments are required: buddy"

    assert message in run(p, "greet howdy --name=John", exit=True)
    assert run(p, "greet howdy John").out == "Howdy John?\n"


# TODO: remove in v.0.30
def test_namespaced_function__deprecated_arg():
    "A subcommand is resolved to a function."

    def hello(name="world"):
        return "Hello {0}!".format(name or "world")

    def howdy(buddy):
        return "Howdy {0}?".format(buddy)

    p = DebugArghParser()
    with pytest.warns(
        DeprecationWarning,
        match=r"Argument `namespace` is deprecated .+ it will be removed in Argh 0\.30",
    ):
        p.add_commands([hello, howdy], namespace="greet")

    assert run(p, "greet hello").out == "Hello world!\n"
    assert run(p, "greet hello --name=John").out == "Hello John!\n"
    assert run(p, "greet hello John", exit=True) == "unrecognized arguments: John"

    # exits with an informative error
    message = "the following arguments are required: buddy"

    assert message in run(p, "greet howdy --name=John", exit=True)
    assert run(p, "greet howdy John").out == "Howdy John?\n"


def test_explicit_cmd_name():
    @argh.named("new-name")
    def orig_name():
        return "ok"

    p = DebugArghParser()
    p.add_commands([orig_name])
    assert "invalid choice" in run(p, "orig-name", exit=True)
    assert run(p, "new-name").out == "ok\n"


def test_aliases():
    @argh.aliases("alias2", "alias3")
    def alias1():
        return "ok"

    p = DebugArghParser()
    p.add_commands([alias1])

    if argh.assembling.SUPPORTS_ALIASES:
        assert run(p, "alias1").out == "ok\n"
        assert run(p, "alias2").out == "ok\n"
        assert run(p, "alias3").out == "ok\n"


def test_help_alias():
    p = DebugArghParser()

    # assert the commands don't fail

    assert run(p, "--help", exit=True) == 0
    assert run(p, "greet --help", exit=True) == 0
    assert run(p, "greet hello --help", exit=True) == 0

    assert run(p, "help", exit=True) == 0
    assert run(p, "help greet", exit=True) == 0
    assert run(p, "help greet hello", exit=True) == 0


def test_arg_order():
    """Positional arguments are resolved in the order in which the @arg
    decorators are defined.
    """

    def cmd(foo, bar):
        return foo, bar

    p = DebugArghParser()
    p.set_default_command(cmd)
    assert run(p, "foo bar").out == "foo\nbar\n"


def test_raw_output():
    "If the raw_output flag is set, no extra whitespace is added"

    def cmd(foo, bar):
        return foo, bar

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "foo bar").out == "foo\nbar\n"
    assert run(p, "foo bar", {"raw_output": True}).out == "foobar"


def test_output_file():
    def cmd():
        return "Hello world!"

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "").out == "Hello world!\n"
    assert run(p, "", {"output_file": None}).out == "Hello world!\n"


def test_command_error():
    def whiner_plain(code=1):
        raise argh.CommandError("I feel depressed.", code=code)

    def whiner_iterable():
        yield "Hello..."
        raise argh.CommandError("I feel depressed.")

    p = DebugArghParser()
    p.add_commands([whiner_plain, whiner_iterable])

    assert run(p, "whiner-plain") == R(
        out="", err="CommandError: I feel depressed.\n", exit=1
    )
    assert run(p, "whiner-plain --code=127") == R(
        out="", err="CommandError: I feel depressed.\n", exit=127
    )
    assert run(p, "whiner-iterable") == R(
        out="Hello...\n", err="CommandError: I feel depressed.\n", exit=1
    )


def test_custom_argparse_namespace():
    @argh.expects_obj
    def cmd(args):
        return args.custom_value

    p = DebugArghParser()
    p.set_default_command(cmd)
    namespace = argparse.Namespace()
    namespace.custom_value = "foo"

    assert run(p, "", {"namespace": namespace}).out == "foo\n"


@pytest.mark.parametrize(
    "argparse_namespace_class", [argparse.Namespace, argh.dispatching.ArghNamespace]
)
def test_get_function_from_namespace_obj(argparse_namespace_class):
    argparse_namespace = argparse_namespace_class()

    def func():
        pass

    retval = argh.dispatching._get_function_from_namespace_obj(argparse_namespace)
    assert retval is None

    setattr(argparse_namespace, argh.constants.DEST_FUNCTION, "")

    retval = argh.dispatching._get_function_from_namespace_obj(argparse_namespace)
    assert retval is None

    setattr(argparse_namespace, argh.constants.DEST_FUNCTION, func)

    retval = argh.dispatching._get_function_from_namespace_obj(argparse_namespace)
    assert retval == func


def test_normalized_keys():
    """Underscores in function args are converted to dashes and back."""

    def cmd(a_b):
        return a_b

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "hello").out == "hello\n"


@mock.patch("argh.assembling.COMPLETION_ENABLED", True)
def test_custom_argument_completer():
    "Issue #33: Enable custom per-argument shell completion"

    @argh.arg("foo", completer="STUB")
    def func(foo):
        pass

    p = argh.ArghParser()
    p.set_default_command(func)

    assert p._actions[-1].completer == "STUB"


def test_class_members():
    "Issue #34: class members as commands"

    class Controller:
        var = 123

        def instance_meth(self, value):
            return value, self.var

        @classmethod
        def class_meth(cls, value):
            return value, cls.var

        @staticmethod
        def static_meth(value):
            return value, "w00t?"

        @staticmethod
        def static_meth2(value):
            return value, "huh!"

    controller = Controller()

    p = DebugArghParser()
    p.add_commands(
        [
            controller.instance_meth,
            controller.class_meth,
            controller.static_meth,
            Controller.static_meth2,
        ]
    )

    assert run(p, "instance-meth foo").out == "foo\n123\n"
    assert run(p, "class-meth foo").out == "foo\n123\n"
    assert run(p, "static-meth foo").out == "foo\nw00t?\n"
    assert run(p, "static-meth2 foo").out == "foo\nhuh!\n"


def test_kwonlyargs():
    "Correct dispatch in presence of keyword-only arguments"

    def cmd(*args, foo="1", bar, baz="3", **kwargs):
        return " ".join(args), foo, bar, baz, len(kwargs)

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "--baz=done test  this --bar=do").out == "test this\n1\ndo\ndone\n0\n"
    message = "the following arguments are required: --bar"
    assert run(p, "test --foo=do", exit=True) == message


def test_default_arg_values_in_help():
    "Argument defaults should appear in the help message implicitly"

    @argh.arg("name", default="Basil")
    @argh.arg("--task", default="hang the Moose")
    @argh.arg("--note", help="why is it a remarkable animal?")
    def remind(
        name,
        task=None,
        reason="there are creatures living in it",
        note="it can speak English",
    ):
        return "Oh what is it now, can't you leave me in peace..."

    p = DebugArghParser()
    p.set_default_command(remind)

    assert "Basil" in p.format_help()
    assert "Moose" in p.format_help()
    assert "creatures" in p.format_help()

    # explicit help message is not obscured by the implicit one...
    assert "remarkable animal" in p.format_help()
    # ...but is still present
    assert "it can speak" in p.format_help()


def test_default_arg_values_in_help__regression():
    "Empty string as default value → empty help string → broken argparse"

    def foo(bar=""):
        return bar

    p = DebugArghParser()
    p.set_default_command(foo)

    # doesn't break
    p.format_help()

    # now check details
    assert "-b BAR, --bar BAR  ''" in p.format_help()
    # note the empty str repr ^^^


def test_help_formatting_is_preserved():
    "Formatting of docstrings should not be messed up in help messages"

    def func():
        """
        Sample function.

        Parameters:
            foo: float
                An example argument.
            bar: bool
                Another argument.
        """
        return "hello"

    p = DebugArghParser()
    p.set_default_command(func)

    assert func.__doc__ in p.format_help()


def test_prog(capsys: pytest.CaptureFixture[str]):
    "Program name propagates from sys.argv[0]"

    def cmd(foo=1):
        return foo

    p = DebugArghParser()
    p.add_commands([cmd])

    usage = get_usage_string()

    assert run(p, "-h", exit=True) == 0
    captured = capsys.readouterr()
    assert captured.out.startswith(usage)


def test_unknown_args():
    def cmd(foo=1):
        return foo

    p = DebugArghParser()
    p.set_default_command(cmd)

    get_usage_string("[-f FOO]")

    assert run(p, "--foo 1") == R(out="1\n", err="")
    assert run(p, "--bar 1", exit=True) == "unrecognized arguments: --bar 1"
    assert run(p, "--bar 1", exit=False, kwargs={"skip_unknown_args": True}) == R(
        out="1\n", err=""
    )


def test_add_commands_no_overrides1(capsys: pytest.CaptureFixture[str]):
    def first_func(foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
    )

    run(p, "--help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp [-h] {{first-func,second-func}} ...

        positional arguments:
          {{first-func,second-func}}
            first-func          Owl stretching time
            second-func

        {HELP_OPTIONS_LABEL}:
          -h, --help            show this help message and exit
        """
        )[1:]
    )


def test_add_commands_no_overrides2(capsys: pytest.CaptureFixture[str]):
    def first_func(foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
    )

    run(p, "first-func --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp first-func [-h] [-f FOO]

        Owl stretching time

        {HELP_OPTIONS_LABEL}:
          -h, --help         show this help message and exit
          -f FOO, --foo FOO  123
        """
        )[1:]
    )


def test_add_commands_group_overrides1(capsys: pytest.CaptureFixture[str]):
    """
    When `group_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        return foo

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
        group_name="my-group",
        group_kwargs={
            "help": "group help override",
            "description": "group description override",
        },
    )

    run(p, "--help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp [-h] {{my-group}} ...

        positional arguments:
          {{my-group}}
            my-group

        {HELP_OPTIONS_LABEL}:
          -h, --help  show this help message and exit
        """
        )[1:]
    )


def test_add_commands_group_overrides2(capsys: pytest.CaptureFixture[str]):
    """
    When `group_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        return foo

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
        group_name="my-group",
        group_kwargs={
            "help": "group help override",
            "description": "group description override",
        },
    )

    run(p, "my-group --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp my-group [-h] {{first-func,second-func}} ...

        {HELP_OPTIONS_LABEL}:
          -h, --help            show this help message and exit

        subcommands:
          group description override

          {{first-func,second-func}}
                                group help override
            first-func          Owl stretching time
            second-func
        """
        )[1:]
    )


def test_add_commands_group_overrides3(capsys: pytest.CaptureFixture[str]):
    """
    When `group_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        return foo

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
        group_name="my-group",
        group_kwargs={
            "help": "group help override",
            "description": "group description override",
        },
    )

    run(p, "my-group first-func --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp my-group first-func [-h] [-f FOO]

        Owl stretching time

        {HELP_OPTIONS_LABEL}:
          -h, --help         show this help message and exit
          -f FOO, --foo FOO  123
        """
        )[1:]
    )


# TODO: remove in v.0.30
def test_add_commands_group_overrides1__deprecated(capsys: pytest.CaptureFixture[str]):
    """
    When `namespace_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    with pytest.warns(DeprecationWarning) as recorded_warnings:
        p.add_commands(
            [first_func, second_func],
            namespace="ns",
            namespace_kwargs={
                "help": "namespace help override",
                "description": "namespace description override",
            },
        )
    assert len(recorded_warnings) == 2
    assert re.match(
        r".*`namespace` is deprecated .+ removed in Argh 0\.30.+ use `group_name` instead",
        str(recorded_warnings[0].message),
    )
    assert re.match(
        r".*`namespace_kwargs` is deprecated .+ removed in Argh 0\.30.+ use `group_kwargs` instead",
        str(recorded_warnings[1].message),
    )

    run(p, "--help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp [-h] {{ns}} ...

        positional arguments:
          {{ns}}
            ns

        {HELP_OPTIONS_LABEL}:
          -h, --help  show this help message and exit
        """
        )[1:]
    )


# TODO: remove in v.0.30
def test_add_commands_group_overrides2__deprecated(capsys: pytest.CaptureFixture[str]):
    """
    When `namespace_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
        namespace="ns",
        namespace_kwargs={
            "help": "namespace help override",
            "description": "namespace description override",
        },
    )

    run(p, "ns --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp ns [-h] {{first-func,second-func}} ...

        {HELP_OPTIONS_LABEL}:
          -h, --help            show this help message and exit

        subcommands:
          namespace description override

          {{first-func,second-func}}
                                namespace help override
            first-func          Owl stretching time
            second-func
        """
        )[1:]
    )


# TODO: remove in v.0.30
def test_add_commands_group_overrides3__deprecated(capsys: pytest.CaptureFixture[str]):
    """
    When `namespace_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
        namespace="ns",
        namespace_kwargs={
            "help": "namespace help override",
            "description": "namespace description override",
        },
    )

    run(p, "ns first-func --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp ns first-func [-h] [-f FOO]

        Owl stretching time

        {HELP_OPTIONS_LABEL}:
          -h, --help         show this help message and exit
          -f FOO, --foo FOO  123
        """
        )[1:]
    )


def test_add_commands_func_overrides1(capsys: pytest.CaptureFixture[str]):
    """
    When `func_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
        func_kwargs={
            "help": "func help override",
            "description": "func description override",
        },
    )

    run(p, "--help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp [-h] {{first-func,second-func}} ...

        positional arguments:
          {{first-func,second-func}}
            first-func          func help override
            second-func         func help override

        {HELP_OPTIONS_LABEL}:
          -h, --help            show this help message and exit
        """
        )[1:]
    )


def test_add_commands_func_overrides2(capsys: pytest.CaptureFixture[str]):
    """
    When `func_kwargs` is passed to `add_commands()`, its members override
    whatever was specified on function level.
    """

    def first_func(foo=123):
        """Owl stretching time"""
        pass

    def second_func():
        pass

    p = argh.ArghParser(prog="myapp")
    p.add_commands(
        [first_func, second_func],
        func_kwargs={
            "help": "func help override",
            "description": "func description override",
        },
    )

    run(p, "first-func --help", exit=True)
    captured = capsys.readouterr()
    assert (
        captured.out
        == unindent(
            f"""
        usage: myapp first-func [-h] [-f FOO]

        func description override

        {HELP_OPTIONS_LABEL}:
          -h, --help         show this help message and exit
          -f FOO, --foo FOO  123
        """
        )[1:]
    )


def test_action_count__only_arg_decorator():
    @argh.arg("-v", "--verbose", action="count", default=0)
    def func(**kwargs):
        verbosity = kwargs.get("verbose")
        return f"verbosity: {verbosity}"

    p = DebugArghParser()
    p.set_default_command(func)

    assert run(p, "").out == "verbosity: 0\n"
    assert run(p, "-v").out == "verbosity: 1\n"
    assert run(p, "-vvvv").out == "verbosity: 4\n"


def test_action_count__mixed():
    @argh.arg("-v", "--verbose", action="count")
    def func(verbose=0):
        return f"verbosity: {verbose}"

    p = DebugArghParser()
    p.set_default_command(func)

    assert run(p, "").out == "verbosity: 0\n"
    assert run(p, "-v").out == "verbosity: 1\n"
    assert run(p, "-vvvv").out == "verbosity: 4\n"



"""
Common stuff for tests
~~~~~~~~~~~~~~~~~~~~~~
"""
import io
import os
import sys
from collections import namedtuple

from argh import ArghParser


# hacky constructor for default exit value
def CmdResult(out, err, exit=None):
    _CmdResult = namedtuple("CmdResult", ("out", "err", "exit"))
    return _CmdResult(out, err, exit)


class DebugArghParser(ArghParser):
    "(does not print stuff to stderr on exit)"

    def exit(self, status=0, message=None):
        raise SystemExit(message)

    def error(self, message):
        self.exit(2, message)


def call_cmd(parser, command_string, **kwargs):
    if hasattr(command_string, "split"):
        args = command_string.split()
    else:
        args = command_string

    io_out = io.StringIO()
    io_err = io.StringIO()

    if "output_file" not in kwargs:
        kwargs["output_file"] = io_out
    kwargs["errors_file"] = io_err

    try:
        result = parser.dispatch(args, **kwargs)
    except SystemExit as e:
        result = None
        exit = e.code or 0  # e.code may be None
    else:
        exit = None

    if kwargs.get("output_file") is None:
        return CmdResult(out=result, err=io_err.read(), exit=exit)
    else:
        io_out.seek(0)
        io_err.seek(0)
        return CmdResult(out=io_out.read(), err=io_err.read(), exit=exit)


def run(parser, command_string, kwargs=None, exit=False):
    """Calls the command and returns CmdResult(out, err, exit status),
    with exit status None if no SystemExit was raised.

    :exit:
        if set to `True` and a SystemExit is raised, the status code is returned;
        if the exception is not raised, an AssertionError is raised.

    """
    kwargs = kwargs or {}
    result = call_cmd(parser, command_string, **kwargs)
    if exit:
        if result.exit is None:
            raise AssertionError("Did not exit")
        else:
            return result.exit
    return result


def get_usage_string(definitions="{cmd} ..."):
    prog = os.path.basename(sys.argv[0])
    return "usage: " + prog + " [-h] " + definitions + "\n\n"



"""
Unit Tests For Assembling Phase
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
from unittest.mock import MagicMock, call, patch

import pytest

import argh


def test_guess_type_from_choices():
    old = dict(option_strings=("foo",), choices=[1, 2])
    new = dict(option_strings=("foo",), choices=[1, 2], type=int)
    assert new == argh.assembling._guess(old)

    # ensure no overrides
    same = dict(option_strings=("foo",), choices=[1, 2], type="NO_MATTER_WHAT")
    assert same == argh.assembling._guess(same)


def test_guess_type_from_default():
    old = dict(option_strings=("foo",), default=1)
    new = dict(option_strings=("foo",), default=1, type=int)
    assert new == argh.assembling._guess(old)

    # ensure no overrides
    same = dict(option_strings=("foo",), default=1, type="NO_MATTER_WHAT")
    assert same == argh.assembling._guess(same)


def test_guess_action_from_default():
    # True → store_false
    old = dict(option_strings=("foo",), default=False)
    new = dict(option_strings=("foo",), default=False, action="store_true")
    assert new == argh.assembling._guess(old)

    # True → store_false
    old = dict(option_strings=("foo",), default=True)
    new = dict(option_strings=("foo",), default=True, action="store_false")
    assert new == argh.assembling._guess(old)

    # ensure no overrides
    same = dict(option_strings=("foo",), default=False, action="NO_MATTER_WHAT")
    assert same == argh.assembling._guess(same)


def test_set_default_command():
    def func():
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        (
            dict(option_strings=("foo",), nargs="+", choices=[1, 2], help="me"),
            dict(
                option_strings=(
                    "-b",
                    "--bar",
                ),
                default=False,
            ),
        ),
    )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    argh.set_default_command(parser, func)

    assert parser.add_argument.mock_calls == [
        call("foo", nargs="+", choices=[1, 2], help="me", type=int),
        call(
            "-b",
            "--bar",
            default=False,
            action="store_true",
            help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE,
        ),
    ]
    assert parser.set_defaults.mock_calls == [call(function=func)]


def test_set_default_command_infer_cli_arg_names_from_func_signature():
    # TODO: split into small tests where we'd check each combo and make sure
    # they interact as expected (e.g. pos opt arg gets the short form even if
    # there's a pos req arg, etc.)
    #
    # an arg with a unique first letter per arg type + every combination
    # of conflicting first letters per every two arg types:
    # - positional required (i.e. without a default value)
    # - positional optional (i.e. with a default value)
    # - named-only required (i.e. kwonly without a default value)
    # - named-only optional (i.e. kwonly with a default valu)
    def func(
        alpha_pos_req,
        beta_pos_req,
        alpha_pos_opt="alpha",
        beta_pos_opt_one="beta one",
        beta_pos_opt_two="beta two",
        gamma_pos_opt="gamma named",
        delta_pos_opt="delta named",
        theta_pos_opt="theta named",
        *args,
        gamma_kwonly_opt="gamma kwonly",
        delta_kwonly_req,
        epsilon_kwonly_req_one,
        epsilon_kwonly_req_two,
        zeta_kwonly_opt="zeta kwonly",
        **kwargs,
    ):
        return (
            alpha_pos_req,
            beta_pos_req,
            alpha_pos_opt,
            beta_pos_opt_one,
            beta_pos_opt_two,
            gamma_pos_opt,
            delta_pos_opt,
            gamma_kwonly_opt,
            delta_kwonly_req,
            epsilon_kwonly_req_one,
            epsilon_kwonly_req_two,
            zeta_kwonly_opt,
            args,
            kwargs,
        )

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()
    parser.set_defaults = MagicMock()

    argh.set_default_command(parser, func)

    help_tmpl = argh.constants.DEFAULT_ARGUMENT_TEMPLATE
    assert parser.add_argument.mock_calls == [
        call("alpha_pos_req", help="%(default)s"),
        call("beta_pos_req", help="%(default)s"),
        call("-a", "--alpha-pos-opt", default="alpha", type=str, help=help_tmpl),
        call("--beta-pos-opt-one", default="beta one", type=str, help=help_tmpl),
        call("--beta-pos-opt-two", default="beta two", type=str, help=help_tmpl),
        call("--gamma-pos-opt", default="gamma named", type=str, help=help_tmpl),
        call("--delta-pos-opt", default="delta named", type=str, help=help_tmpl),
        call("-t", "--theta-pos-opt", default="theta named", type=str, help=help_tmpl),
        call("--gamma-kwonly-opt", default="gamma kwonly", type=str, help=help_tmpl),
        call("--delta-kwonly-req", required=True, help=help_tmpl),
        call("--epsilon-kwonly-req-one", required=True, help=help_tmpl),
        call("--epsilon-kwonly-req-two", required=True, help=help_tmpl),
        call(
            "-z", "--zeta-kwonly-opt", default="zeta kwonly", type=str, help=help_tmpl
        ),
        call("args", nargs="*", help=help_tmpl),
    ]
    assert parser.set_defaults.mock_calls == [call(function=func)]


def test_set_default_command_docstring():
    def func():
        "docstring"
        pass

    parser = argh.ArghParser()

    argh.set_default_command(parser, func)

    assert parser.description == "docstring"


def test_add_subparsers_when_default_command_exists():
    def one():
        return 1

    def two():
        return 2

    def three():
        return 3

    p = argh.ArghParser()
    p.set_default_command(one)
    p.add_commands([two, three])

    ns_default = p.parse_args([])

    ns_explicit_two = p.parse_args(["two"])
    ns_explicit_three = p.parse_args(["three"])

    assert ns_default.get_function() == one
    assert ns_explicit_two.get_function() == two
    assert ns_explicit_three.get_function() == three


def test_set_default_command_when_subparsers_exist():
    def one():
        return 1

    def two():
        return 2

    def three():
        return 3

    p = argh.ArghParser()
    p.add_commands([one, two])
    p.set_default_command(three)

    ns_default = p.parse_args([])
    ns_explicit_one = p.parse_args(["one"])
    ns_explicit_two = p.parse_args(["two"])

    assert ns_explicit_one.get_function() == one
    assert ns_explicit_two.get_function() == two
    assert ns_default.get_function() == three


def test_add_command_with_group_kwargs_but_no_group_name():
    def one():
        return 1

    p = argh.ArghParser()
    err_msg = "`group_kwargs` only makes sense with `group_name`"
    with pytest.raises(ValueError, match=err_msg):
        p.add_commands([one], group_kwargs={"help": "foo"})


def test_set_default_command_mixed_arg_types():
    def func():
        pass

    setattr(func, argh.constants.ATTR_ARGS, [dict(option_strings=("x", "--y"))])

    p = argh.ArghParser()

    with pytest.raises(argh.AssemblingError) as excinfo:
        p.set_default_command(func)
    msg = "func: cannot add x/--y: invalid option string"
    assert msg in str(excinfo.value)


def test_set_default_command_varargs():
    def func(*file_paths):
        yield ", ".join(file_paths)

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()

    argh.set_default_command(parser, func)

    assert parser.add_argument.mock_calls == [
        call("file_paths", nargs="*", help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
    ]


def test_set_default_command_kwargs():
    @argh.arg("foo")
    @argh.arg("--bar")
    def func(x, **kwargs):
        pass

    parser = argh.ArghParser()

    parser.add_argument = MagicMock()

    argh.set_default_command(parser, func)

    assert parser.add_argument.mock_calls == [
        call("x", help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
        call("foo", help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
        call("--bar", help=argh.constants.DEFAULT_ARGUMENT_TEMPLATE),
    ]


# TODO: remove in v.0.30
def test_annotation():
    "Extracting argument help from function annotations."

    def cmd(foo: "quux" = 123):  # noqa: F821
        pass

    parser = argh.ArghParser()

    with pytest.warns(DeprecationWarning, match="will be removed in Argh 0.30"):
        parser.set_default_command(cmd)

    prog_help = parser.format_help()

    assert "quux" in prog_help


def test_kwonlyargs():
    "Correctly processing required and optional keyword-only arguments"

    def cmd(foo_pos, bar_pos, *args, foo_kwonly="foo_kwonly", bar_kwonly):
        return (foo_pos, bar_pos, args, foo_kwonly, bar_kwonly)

    p = argh.ArghParser()
    p.add_argument = MagicMock()
    p.set_default_command(cmd)
    help_tmpl = argh.constants.DEFAULT_ARGUMENT_TEMPLATE
    assert p.add_argument.mock_calls == [
        call("foo_pos", help=help_tmpl),
        call("bar_pos", help=help_tmpl),
        call("-f", "--foo-kwonly", default="foo_kwonly", type=str, help=help_tmpl),
        call("-b", "--bar-kwonly", required=True, help=help_tmpl),
        call("args", nargs="*", help=help_tmpl),
    ]


@patch("argh.assembling.COMPLETION_ENABLED", True)
def test_custom_argument_completer():
    "Issue #33: Enable custom per-argument shell completion"

    def func(foo):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [dict(option_strings=("foo",), completer="STUB")],
    )

    p = argh.ArghParser()
    p.set_default_command(func)

    assert p._actions[-1].completer == "STUB"


@patch("argh.assembling.COMPLETION_ENABLED", False)
def test_custom_argument_completer_no_backend():
    "If completion backend is not available, nothing breaks"

    def func(foo):
        pass

    setattr(
        func,
        argh.constants.ATTR_ARGS,
        [dict(option_strings=("foo",), completer="STUB")],
    )

    p = argh.ArghParser()
    p.set_default_command(func)

    assert not hasattr(p._actions[-1], "completer")


# TODO: remove in v.0.30
def test_set_default_command_deprecation_warnings():
    parser = argh.ArghParser()

    with pytest.warns(
        DeprecationWarning, match="Argument `title` is deprecated in add_commands()"
    ):
        argh.add_commands(parser, [], group_name="a", title="bar")

    with pytest.warns(
        DeprecationWarning,
        match="Argument `description` is deprecated in add_commands()",
    ):
        argh.add_commands(parser, [], group_name="b", description="bar")

    with pytest.warns(
        DeprecationWarning, match="Argument `help` is deprecated in add_commands()"
    ):
        argh.add_commands(parser, [], group_name="c", help="bar")


@patch("argh.assembling.add_commands")
def test_add_subcommands(mock_add_commands):
    mock_parser = MagicMock()

    def get_items():
        pass

    argh.add_subcommands(
        mock_parser,
        "db",
        [get_items],
        title="database commands",
        help="CRUD for our silly database",
    )

    mock_add_commands.assert_called_with(
        mock_parser,
        [get_items],
        group_name="db",
        group_kwargs={
            "title": "database commands",
            "help": "CRUD for our silly database",
        },
    )


@patch("argh.helpers.autocomplete")
def test_arghparser_autocomplete_method(mock_autocomplete):
    p = argh.ArghParser()
    p.autocomplete()

    mock_autocomplete.assert_called()


def test_is_positional():
    with pytest.raises(ValueError, match="Expected at least one"):
        argh.assembling._is_positional([])
    with pytest.raises(ValueError, match="Expected at least one"):
        argh.assembling._is_positional([""])
    assert argh.assembling._is_positional(["f"]) is True
    assert argh.assembling._is_positional(["foo"]) is True
    assert argh.assembling._is_positional(["--foo"]) is False
    assert argh.assembling._is_positional(["-f"]) is False
    assert argh.assembling._is_positional(["-f", "--foo"]) is False

    # this spec is invalid but validation is out of scope of the function
    # as it only checks if the first argument has the leading dash
    assert argh.assembling._is_positional(["-f", "foo"]) is False



"""
Dispatching tests
~~~~~~~~~~~~~~~~~
"""
import io
from unittest.mock import Mock, patch

import pytest

import argh


def _dispatch_and_capture(func, command_string, **kwargs):
    if hasattr(command_string, "split"):
        args = command_string.split()
    else:
        args = command_string

    _io = io.StringIO()
    if "output_file" not in kwargs:
        kwargs["output_file"] = _io

    result = argh.dispatch_command(func, args, **kwargs)

    if kwargs.get("output_file") is None:
        return result
    else:
        _io.seek(0)
        return _io.read()


def run_func(func, command_string, **kwargs):
    try:
        result = _dispatch_and_capture(func, command_string, **kwargs)
    except SystemExit:
        raise
    else:
        return result


@patch("argh.dispatching.argparse.ArgumentParser")
@patch("argh.dispatching.dispatch")
@patch("argh.dispatching.set_default_command")
def test_dispatch_command(mock_set_default_command, mock_dispatch, mock_parser_class):
    def func():
        pass

    argh.dispatching.dispatch_command(func)

    mock_parser_class.assert_called_once()
    mock_parser = mock_parser_class.return_value
    mock_set_default_command.assert_called_with(mock_parser, func)
    mock_dispatch.assert_called_with(mock_parser)


@patch("argh.dispatching.argparse.ArgumentParser")
@patch("argh.dispatching.dispatch")
@patch("argh.dispatching.add_commands")
def test_dispatch_commands(mock_add_commands, mock_dispatch, mock_parser_class):
    def func():
        pass

    argh.dispatching.dispatch_commands([func])

    mock_parser_class.assert_called_once()
    mock_parser = mock_parser_class.return_value
    mock_add_commands.assert_called_with(mock_parser, [func])
    mock_dispatch.assert_called_with(mock_parser)


@patch("argh.dispatching.dispatch")
@patch("argh.dispatching.add_commands")
@patch("argparse.ArgumentParser")
def test_entrypoint(ap_cls_mock, add_commands_mock, dispatch_mock):

    entrypoint = argh.EntryPoint("my cool app")

    # no commands

    with pytest.raises(argh.exceptions.DispatchingError) as excinfo:
        entrypoint()
    assert excinfo.exconly().endswith(
        'DispatchingError: no commands for entry point "my cool app"'
    )

    mocked_parser = Mock()
    ap_cls_mock.return_value = mocked_parser

    # a single command

    @entrypoint
    def greet():
        return "hello"

    entrypoint()

    assert add_commands_mock.called
    add_commands_mock.assert_called_with(mocked_parser, [greet])
    assert dispatch_mock.called
    dispatch_mock.assert_called_with(mocked_parser)

    # multiple commands

    add_commands_mock.reset_mock()
    dispatch_mock.reset_mock()

    @entrypoint
    def hit():
        return "knight with a chicken"

    entrypoint()

    assert add_commands_mock.called
    add_commands_mock.assert_called_with(mocked_parser, [greet, hit])
    assert dispatch_mock.called
    dispatch_mock.assert_called_with(mocked_parser)



"""
Interaction Tests
~~~~~~~~~~~~~~~~~
"""
import unittest.mock as mock

import argh


def parse_choice(choice, **kwargs):
    with mock.patch("argh.interaction.input", lambda prompt: choice):
        return argh.confirm("test", **kwargs)


def test_simple():
    assert parse_choice("") is None
    assert parse_choice("", default=None) is None
    assert parse_choice("", default=True) is True
    assert parse_choice("", default=False) is False

    assert parse_choice("y") is True
    assert parse_choice("y", default=True) is True
    assert parse_choice("y", default=False) is True
    assert parse_choice("y", default=None) is True

    assert parse_choice("n") is False
    assert parse_choice("n", default=True) is False
    assert parse_choice("n", default=False) is False
    assert parse_choice("n", default=None) is False

    assert parse_choice("x") is None


def test_prompt():
    "Prompt is properly formatted"
    prompts = []

    def raw_input_mock(prompt):
        prompts.append(prompt)

    with mock.patch("argh.interaction.input", raw_input_mock):
        argh.confirm("do smth")
        assert prompts[-1] == "do smth? (y/n)"

        argh.confirm("do smth", default=None)
        assert prompts[-1] == "do smth? (y/n)"

        argh.confirm("do smth", default=True)
        assert prompts[-1] == "do smth? (Y/n)"

        argh.confirm("do smth", default=False)
        assert prompts[-1] == "do smth? (y/N)"


@mock.patch("argh.interaction.input")
def test_encoding(mock_input):
    "Unicode is accepted as prompt message"

    msg = "привет"

    argh.confirm(msg)

    mock_input.assert_called_once_with("привет? (y/n)")


@mock.patch("argh.interaction.input")
def test_skip(mock_input):
    retval = argh.confirm("test", default=123, skip=True)

    assert retval == 123
    mock_input.assert_not_called()


@mock.patch("argh.interaction.input")
def test_keyboard_interrupt(mock_input):
    mock_input.side_effect = KeyboardInterrupt
    retval = argh.confirm("test")
    assert retval is None



"""
Regression tests
~~~~~~~~~~~~~~~~
"""
import pytest

import argh

from .base import DebugArghParser, run


def test_regression_issue12():
    """
    Issue #12: @command was broken if there were more than one argument
    to begin with same character (i.e. short option names were inferred
    incorrectly).
    """

    def cmd(foo=1, fox=2):
        yield "foo {0}, fox {1}".format(foo, fox)

    p = DebugArghParser()
    p.set_default_command(cmd)

    assert run(p, "").out == "foo 1, fox 2\n"
    assert run(p, "--foo 3").out == "foo 3, fox 2\n"
    assert run(p, "--fox 3").out == "foo 1, fox 3\n"
    assert "unrecognized" in run(p, "-f 3", exit=True)


def test_regression_issue12_help_flag():
    """
    Issue #12: if an argument starts with "h", e.g. "--host",
    ArgumentError is raised because "--help" is always added by argh
    without decorators.
    """

    def ddos(host="localhost"):
        return "so be it, {0}!".format(host)

    # no help → no conflict
    p = DebugArghParser("PROG", add_help=False)
    p.set_default_command(ddos)
    assert run(p, "-h 127.0.0.1").out == "so be it, 127.0.0.1!\n"

    # help added → conflict → short name ignored
    p = DebugArghParser("PROG", add_help=True)
    p.set_default_command(ddos)
    assert run(p, "-h 127.0.0.1", exit=True) == 0


def test_regression_issue27():
    """
    Issue #27: store_true is not set for inferred bool argument.

    :Reason: when @command was refactored, it stopped using @arg, but it is
    it was there that guesses (choices→type, default→type and
    default→action) were made.
    """

    def parrot(dead=False):
        return "this parrot is no more" if dead else "beautiful plumage"

    def grenade(count=3):
        if count == 3:
            return "Three shall be the number thou shalt count"
        else:
            return "{0!r} is right out".format(count)

    p = DebugArghParser()
    p.add_commands([parrot, grenade])

    # default → type (int)
    assert run(p, "grenade").out == ("Three shall be the number " "thou shalt count\n")
    assert run(p, "grenade --count 5").out == "5 is right out\n"

    # default → action (store_true)
    assert run(p, "parrot").out == "beautiful plumage\n"
    assert run(p, "parrot --dead").out == "this parrot is no more\n"


def test_regression_issue31():
    """
    Issue #31: Argh fails with parameter action type 'count' if a default
    value is provided.

    :Reason: assembling._guess() would infer type from default value without
        regard to the action.  _CountAction does not accept argument "type".

    :Solution: restricted type inferring to actions "store" and "append".
    """

    @argh.arg("-v", "--verbose", dest="verbose", action="count", default=0)
    def cmd(**kwargs):
        yield kwargs.get("verbose", -1)

    p = DebugArghParser()
    p.set_default_command(cmd)
    assert "0\n" == run(p, "").out
    assert "1\n" == run(p, "-v").out
    assert "2\n" == run(p, "-vv").out


def test_regression_issue47():
    @argh.arg("--foo-bar", default="full")
    def func(foo_bar):
        return "hello"

    p = DebugArghParser()
    with pytest.raises(argh.assembling.AssemblingError) as excinfo:
        p.set_default_command(func)
    msg = (
        'func: argument "foo_bar" declared as positional (in function '
        "signature) and optional (via decorator)"
    )
    assert excinfo.exconly().endswith(msg)


def test_regression_issue76():
    """
    Issue #76: optional arguments defaulting to the empty string break --help.

    This is also tested in integration tests but in a different way.
    """

    def cmd(foo=""):
        pass

    p = DebugArghParser()
    p.set_default_command(cmd)
    run(p, "--help", exit=True)


def test_regression_issue104():
    """
    Issue #76: Bug in the way **kwargs is handled

    **kwargs handling was broken in the case that required (no default
    value) positional argument names contained underscores.
    """

    def cmd(foo_foo, bar_bar, baz_baz=5, bip_bip=9, **kwargs):
        return "\n".join(
            [str(foo_foo), str(bar_bar), str(baz_baz), str(bip_bip), str(kwargs)]
        )

    p = DebugArghParser()
    p.set_default_command(cmd)
    expected = "abc\ndef\n8\n9\n{}\n"
    assert run(p, "abc def --baz-baz 8").out == expected
